from __future__ import annotations

import functools
import logging.config
from abc import ABC
from types import SimpleNamespace
from typing import Optional
from typing import Protocol
from typing import Type
from typing import Union

from freshdesk.enumerators import HTTPRequestMethod
from freshdesk.logger import get_config


""" Logging """
logging.config.dictConfig(get_config(__file__))

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def register_interface(function):
    function.register = True

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        function.register = True

        return function(*args, **kwargs)

    return wrapper


class API(Protocol):
    name: Optional[str] = None
    client: Optional[Type[Client]] = None

    @classmethod
    def _init_client(cls, client: Type[Client] = None) -> None:
        cls.client = client
        ...

    def _register_interfaces(self) -> None: ...


class Client(Protocol):
    api: SimpleNamespace

    def __init__(self):
        self._initialize_apis()
        ...

    def _initialize_apis(self) -> None:
        for api in self.api.__dict__:
            APIClass = getattr(self.api, api)
            api_instance = APIClass(self)

            setattr(self.api, api, api_instance)
        ...

    @classmethod
    def register_api(cls, api: Type[API]) -> None: ...


class BaseClient(ABC):
    api: SimpleNamespace = SimpleNamespace()
    _initialized: bool = False

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        if not cls._initialized:
            instance._initialize_apis()
            cls._initialized = True
        return instance

    def _initialize_apis(self) -> None:
        for api in self.api.__dict__:
            APIClass = getattr(self.api, api)
            api_instance = APIClass(self)

            api_instance.client = self
            setattr(self.api, api, api_instance)

    @classmethod
    def register_api(cls, api: Type[API]) -> None:
        # Init Client with API
        api._init_client(cls)

        # Add API to self.api
        api_name = getattr(api, "name", api.__name__)  # type: ignore
        setattr(cls.api, api_name, api)


class BaseAPI(ABC):
    client = None
    path = "/"

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance.client = getattr(instance, "client", None)
        instance._register_interfaces()
        return instance

    @classmethod
    def _init_client(cls, client):
        cls.client = client

    @property
    def base_url(self):
        if self.client is not None:
            return self.client.base_url + self.path

        return "https://domain.freshdesk.com/"

    def _register_interfaces(self) -> None:
        # Register Intefaces to Client
        for function_name in (f for f in dir(self) if not f.startswith("_")):
            try:
                function = getattr(self, function_name)
            except Exception:  # Hotfix for errors with properties
                continue
            register = getattr(function, "register", False)

            if register:
                setattr(self.client, function.__name__, function)

    def _request(
        self, url: str, method: Union[str, HTTPRequestMethod] = "GET", **kwargs
    ):
        return self.client._request(url=url, method=method, **kwargs)



from __future__ import annotations

import enum
import inspect
import logging.config
from copy import deepcopy
from dataclasses import dataclass
from dataclasses import field
from datetime import date
from datetime import datetime
from datetime import timezone
from enum import Enum
from enum import IntEnum
from pathlib import Path
from typing import Any
from typing import Optional
from typing import TypedDict
from typing import Union
from urllib.parse import ParseResult

from freshdesk.enumerators import LogicalOperator
from freshdesk.enumerators import SolutionArticleStatus
from freshdesk.enumerators import SolutionFolderVisibility
from freshdesk.logger import get_config
from freshdesk.typings import TimeDeltaString


# Logging
logging.config.dictConfig(get_config(__file__))

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@dataclass
class Attachment:
    id: int
    name: Path
    content_type: str  # MIME Type
    size: int
    created_at: datetime
    updated_at: datetime
    attachment_url: ParseResult

    # catch any additional fields
    extras: dict[str, Any] = field(default_factory=dict)

    _json: Optional[dict] = field(default=None, repr=False)

    @classmethod
    def from_json(cls, data: dict) -> Attachment:
        obj = cls(
            id=data["id"],
            name=Path(data["name"]),
            content_type=data["content_type"],
            size=data["size"],
            created_at=datetime.fromisoformat(data["created_at"][:-1]),
            updated_at=datetime.fromisoformat(data["updated_at"][:-1]),
            attachment_url=data["attachment_url"],
            extras={
                key: value
                for key, value in data.items()
                if key not in cls.__dataclass_fields__
            },  # noqa: 501
        )
        obj._json = data

        return obj

    def to_json(self) -> dict:
        if self._json:
            return self._json

        return {
            "id": self.id,
            "name": str(self.name),
            "content_type": self.content_type,
            "size": self.size,
            "created_at": self.created_at.isoformat()[:-6] + "Z",
            "updated_at": self.updated_at.isoformat()[:-6] + "Z",
            "attachment_url": self.attachment_url,
        }


class TicketStatus(IntEnum):
    """The status of a ticket."""

    OPEN = 2
    PENDING = 3
    RESOLVED = 4
    CLOSED = 5
    OTHER = 6


@dataclass
class Ticket:
    # Primary Key
    id: int

    # Ticket Properties
    type: str = ""
    source: Optional[int] = None
    status: Optional[TicketStatus] = None
    tags: list[str] = field(default_factory=list)
    priority: Optional[int] = None
    deleted: bool = False
    custom_fields: dict = field(default_factory=dict)

    # User/Email
    name: str = ""
    phone: str = ""
    email: str = ""

    support_email: Optional[str] = None
    associated_tickets_count: Optional[int] = None

    nr_due_by: Optional[datetime] = None
    nr_escalated: bool = False

    subject: str = ""
    attachments: list[Attachment] = field(default_factory=list)
    to_emails: list[str] = field(default_factory=list)
    cc_emails: list[str] = field(default_factory=list)
    ticket_cc_emails: list[str] = field(default_factory=list)
    reply_cc_emails: list[str] = field(default_factory=list)
    fwd_emails: list[str] = field(default_factory=list)
    description: str = ""
    description_text: str = ""

    # Foreign Keys
    email_config_id: Optional[int] = None

    company_id: Optional[int] = None
    group_id: Optional[int] = None
    product_id: Optional[int] = None
    requester_id: Optional[int] = None
    responder_id: Optional[int] = None

    twitter_id: str = ""
    facebook_id: str = ""
    form_id: Optional[int] = None

    # Datetime
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    due_by: Optional[datetime] = None

    fr_due_by: Optional[datetime] = None

    # Bools
    fr_escalated: bool = False
    is_escalated: bool = False
    spam: bool = False

    # Other Fields
    urgent: bool = False
    association_type: Optional[int] = None
    # List of ids for associated tickets
    associated_tickets_list: list[int] = field(default_factory=list)
    source_additional_info: Optional[None] = None

    _json: Optional[dict] = field(default=None, repr=False)

    # Any extra arguments should be put in here.
    extras: dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_json(cls, data) -> Ticket:
        copied = deepcopy(data)

        # Extras
        copied["extras"] = {}
        for key in list(copied.keys()):
            if key not in cls.__dataclass_fields__:
                copied["extras"][key] = copied.pop(key)

        obj = cls(**copied)

        # Parse datetime strings into datetime objects
        for attr in ["created_at", "updated_at", "due_by", "fr_due_by"]:
            if copied.get(attr):
                setattr(obj, attr, datetime.fromisoformat(copied[attr][:-1]))
        # Parse Status
        if copied.get("status"):
            obj.status = TicketStatus(copied["status"])
        # Parse Attachments
        if copied.get("attachments"):
            obj.attachments = [
                Attachment.from_json(a) for a in copied["attachments"]
            ]

        obj._json = data

        return obj

    def to_json(self):
        raise NotImplementedError(f"{self} is not yet JSON serializable.")

    def __hash__(self):
        return hash(self.id)


@dataclass
class TicketConversation:
    """Conversational item for a ticket. Can be a Note or Reply."""

    id: int
    ticket_id: int
    user_id: int
    thread_id: Optional[int]
    thread_message_id: Optional[int]
    automation_id: Optional[int]
    automation_type_id: Optional[int]
    source: int
    source_additional_info: None  # What is this?
    category: int  # What is this?

    body: str
    body_text: str

    to_emails: list[str]
    cc_emails: list[str]
    bcc_emails: Optional[list[str]]
    from_email: Optional[str]
    support_email: Optional[str]

    attachments: list[Attachment]

    incoming: bool
    private: bool
    auto_response: bool

    email_failure_count: Optional[int]
    outgoing_failures: Optional[int]

    created_at: datetime
    updated_at: datetime
    last_edited_at: Optional[datetime]
    last_edited_user_id: Optional[int]

    extras: dict[str, Any] = field(default_factory=dict)
    _json: Optional[dict] = field(default=None, repr=False)

    @classmethod
    def from_json(cls, data):
        copied = deepcopy(data)

        # Parse Datetimes
        for key in ["created_at", "updated_at", "last_edited_at"]:
            if copied.get(key):
                copied[key] = datetime.fromisoformat(copied[key][:-1])

        # Parse Attachments
        if copied.get("attachments"):
            copied['attachments'] = [
                Attachment.from_json(attachment)
                for attachment in copied["attachments"]
            ]

        # Extras
        extras = {}
        for key in list(copied.keys()):
            if key not in cls.__dataclass_fields__:
                extras[key] = copied.pop(key)

        return cls(
            **copied,
            extras=extras,
            _json=data,
        )


@dataclass
class CannedResponseFolder:
    """A Freshdesk _Folder_ of Canned Responses."""

    id: int
    name: str
    created_at: Optional[datetime]
    updated_at: Optional[datetime]

    personal: Optional[bool] = None
    responses_count: Optional[int] = None
    canned_responses: Optional[list[CannedResponse]] = field(
        default=None, repr=False
    )

    _json: Optional[dict] = field(default=None, repr=False)

    def __len__(self) -> int:
        return self.responses_count if self.responses_count is not None else len(self.canned_responses)  # type: ignore

    @classmethod
    def from_json(cls, data) -> CannedResponseFolder:
        copied = deepcopy(data)

        # Parse Datetimes
        for key in ("created_at", "updated_at"):
            if copied.get(key):
                copied[key] = datetime.fromisoformat(copied[key][:-1])

        obj = cls(**copied)
        obj._json = data

        if copied.get("canned_responses"):
            obj.canned_responses = [
                CannedResponse.from_json(canned_response)
                for canned_response in copied.get("canned_responses")
            ]

        return obj

    def to_json(self) -> dict:
        if self._json:
            return self._json

        crs = None
        if self.canned_responses:
            crs = [cr.to_json() for cr in self.canned_responses]
        return {
            "id": self.id,
            "name": self.name,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "personal": self.personal,
            "responses_count": self.responses_count,
            "canned_responses": crs,
        }


class CannedResponseVisibility(enum.IntEnum):
    """Visibility of a Canned Response."""

    ALL_AGENTS = 0
    PERSONAL = 1
    SELECT_GROUPS = 2


@dataclass
class CannedResponse:
    """Canned Response in Freshdesk.

    This is like a code snippet, but for a support ticket.

    There's a simple version, supplied by the Canned Response Folder API that
    only provides the ID and title.
    """

    id: int
    title: str

    content: str = field(repr=False)
    content_html: str = field(repr=False)

    responses_count: Optional[int] = None
    folder_id: Optional[int] = None
    # Groups for which the canned response is visible. Use only if visibility
    # is set to 2.
    group_ids: Optional[list[int]] = None
    visibility: Optional[CannedResponseVisibility] = None

    personal: Optional[bool] = None

    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    attachments: list[Attachment] = field(default_factory=list)

    _json: Optional[dict] = field(default=None, repr=False)

    @classmethod
    def from_json(cls, data: dict):
        copied = deepcopy(data)

        # Parse Datetimes
        for key in ["created_at", "updated_at"]:
            if copied.get(key):
                copied[key] = datetime.fromisoformat(copied[key][:-1]).astimezone(
                    timezone.utc
                )

        if copied.get("visibility"):
            copied["visibility"] = CannedResponseVisibility(copied["visibility"])

        if copied.get("attachments"):
            copied["attachments"] = [
                Attachment.from_json(attachment)
                for attachment in copied.get("attachments", [])
            ]

        obj = cls(**copied)
        obj._json = data
        return obj

    def to_json(self) -> dict:
        if self._json:
            return self._json

        attachments = None
        if self.attachments:
            attachments = [
                attachment.to_json() for attachment in self.attachments
            ]

        return {
            "id": self.id,
            "title": self.title,
            "content": self.content,
            "content_html": self.content_html,
            "responses_count": self.responses_count,
            "folder_id": self.folder_id,
            "group_ids": self.group_ids,
            "visibility": self.visibility,
            "personal": self.personal,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "attachments": attachments,
        }


@dataclass
class TicketField:
    id: int
    name: str
    type: Optional[str] = None
    description: str = ""
    label: Optional[str] = None
    position: Optional[int] = None

    portal_cc: Optional[bool] = None
    portal_cc_to: Optional[list[str]] = None

    choices: Optional[list[str]] = None
    dependent_fields: None = None
    section_mappings: None = None

    default: Optional[bool] = None
    archived: Optional[bool] = None
    has_section: Optional[bool] = None
    is_fsm: Optional[bool] = None
    field_update_in_progress: Optional[bool] = None

    displayed_to_customers: Optional[bool] = None
    customers_can_edit: Optional[bool] = None
    label_for_customers: Optional[str] = None

    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    required_for_closure: bool = False
    required_for_agents: bool = False
    required_for_customers: bool = False

    def __post_init__(self):
        signature = inspect.signature(self.__init__)
        default_values = {
            key: value.default for key, value in signature.parameters.items()
        }

        logger.info(
            f'{self.__class__.__name__} __init__() missing {len(default_values)} required positional arguments: {", ".join(default_values.keys())}'
        )  # noqa: 501

        for key, value in default_values.items():
            if getattr(self, key) is None:
                logger.info(
                    f"{self.__class__.__name__} __init__() has None for {key} with default value {value}."  # noqa: 501
                )


CustomFieldName = str
CustomFieldValue = Union[str, int, bool]


@dataclass
class TicketFilter:
    """

    [Documentation](https://developers.freshdesk.com/api/#filter_tickets)


    Paramaters
    -----
    custom_fields: Optional[dict[CustomFieldName, CustomFieldValue]]
        Single line text: string
        Number: integer
        Checkbox: boolean
        Dropdown: string

    Specs
    -----

    * Archived tickets are not included.
    * Dates: YYYY-MM-DD (UTC)
    * Must be URL Encoded - e.g. ' ' -> '%20'
    * <= 512 characters, including surrounding double quotes

    Types
    =====

    * null
    * string
    * integer
    * boolean
    * date (YYYY-MM-DD)
        * surrounded by single quotes
        * e.g. '2018-01-01'

    Operators
    =========

    Logical Operators

    * `AND`
    * `OR`
    * `(...)`

    Relational Operators

    `:`
    : Equals

    `:>`
    : Greater than
    : date and numeric ONLY

    `:<`
    : Less than
    : date and numeric ONLY

    Returns
    -------

    Up to 30 tickets with appropriate pagination infromation

    Prototype
    ---------

    >>> from pyfreshdesk.filters import Field

    >>> ticket_filter = TicketFilter(...)

    >>> field_1 = Field(agent_id=None)
    "agent_id:null"

    >>> field_1 = Field(agent_id=1)
    >>> field_2 = Field(created_at='2021-11-09')
    >>>
    >>> field_1 & field_2
    "agent_id:1 AND created_at:'2021-11-09'"

    >>> field_1 = Field(agent_id=1)
    >>> field_2 = Field(created_at='2021-11-09')
    >>> field_3 = Field(created_at='2020-09-08')
    >>>
    >>> Parenthesize(field_1 & field_2) | field_3
    "(agent_id:1 AND created_at:'2021-11-09') OR created_at:'2020-09-08'"

    """

    agent_id: Optional[int] = None
    group_id: Optional[int] = None
    priority: Optional[int] = None
    status: Optional[int] = None
    tag: Optional[str] = None
    type: Optional[str] = None
    due_by: Optional[date] = None
    fr_due_by: Optional[date] = None
    created_at: Optional[date] = None
    updated_at: Optional[date] = None

    custom_fields: Optional[dict[CustomFieldName, CustomFieldValue]] = None


class Field:
    """A single field to filter on."""

    DATE_FORMAT = "%Y-%m-%d"

    supported_types = (type(None), str, int, bool, date)

    builtin_field_types = {
        "agent_id": int,
        "group_id": int,
        "priority": int,
        "status": int,
        "tag": str,
        "type": str,
        "due_by": date,
        "fr_due_by": date,
        "created_at": date,
        "updated_at": date,
    }

    def __init__(self, **kwargs):
        self._validate_kwargs(kwargs)

        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        # agent_id:1
        kwargs = {
            key: getattr(self, key)
            for key in vars(self)
            if not key.startswith("_")
        }
        self._validate_kwargs(kwargs)

        key = list(kwargs.keys())[0]
        value = self._resolve_value(list(kwargs.values())[0])

        return f"{key}:{value}"

    def __and__(self, other):
        return f"{self!s} {LogicalOperator.AND!s} {other!s}"

    def __or__(self, other):
        return f"{self!s} {LogicalOperator.OR!s} {other!s}"

    def _validate_kwargs(self, kwargs):
        # 1 key:value pair only
        self._validate_kwargs_length(kwargs)
        self._validate_kwargs_type(kwargs)

        # Validate Type

    def _validate_kwargs_length(self, kwargs):
        # Specific Supported Types
        if len(kwargs) > 1 or len(kwargs) == 0:
            raise ValueError(
                f"{self.__class__.__name__} requires one field. Got {len(kwargs)}."  # noqa: 501
            )

    def _validate_kwargs_type(self, kwargs):
        logger.info(f"Validating kwargs type: {kwargs=}.")
        key = list(kwargs.keys())[0]
        value = list(kwargs.values())[0]

        # All Supported Types
        if not isinstance(value, self.supported_types):
            raise TypeError(
                f"{self.__class__.__name__} requires a string, integer, boolean, or date. Got {type(value)}."  # noqa: 501
            )

        # Built-In Supported Types
        if key in self.builtin_field_types:
            if not isinstance(
                value, (self.builtin_field_types[key], type(None))
            ):
                raise TypeError(
                    f"{self.__class__.__name__} requires a {self.builtin_field_types[key]}. Got {type(value)}."  # noqa: 501
                )

    def _resolve_value(self, value):
        """Will try to convert value to appropriate format/type."""

        if value is None:
            return "null"

        if isinstance(value, date):
            return "'" + value.strftime(self.DATE_FORMAT) + "'"

        return value


class Parenthesize:
    def __call__(self, arguments):
        return f"({arguments})"


@dataclass
class Product:
    """A Freshdesk Product."""

    id: int
    name: str
    description: str
    primary_email: str
    created_at: date
    updated_at: date

    @classmethod
    def from_json(cls, data):
        obj = cls(**data)
        return obj


@dataclass
class Group:
    """A Freshdesk Group."""

    id: int
    business_hour_id: int
    name: str  # unique
    created_at: date
    updated_at: date

    group_type: str = ""
    agent_ids: list[int] = field(default_factory=list)
    auto_ticket_assign: int = 0
    description: str = ""
    escalate_to: Optional[int] = None

    unassigned_for: TimeDeltaString = ""

    extras: dict[str, Any] = field(default_factory=dict)
    _json: Optional[dict] = field(default=None, repr=False)

    @classmethod
    def from_json(cls, data):
        copied = deepcopy(data)

        # Parse Datetimes
        copied["created_at"] = datetime.fromisoformat(
            copied["created_at"][:-1]
        ).astimezone(timezone.utc)
        copied["updated_at"] = datetime.fromisoformat(
            copied["updated_at"][:-1]
        ).astimezone(timezone.utc)

        # Extras
        extras = {}
        for key in list(copied.keys()):
            if key not in cls.__dataclass_fields__:
                extras[key] = copied.pop(key)

        obj = cls(**copied, extras=extras, _json=data)

        return obj


class SolutionCategoryPayload(TypedDict):
    """Used to create or update a solution category using the API."""

    name: str
    description: Optional[str]
    visible_in_portals: Optional[list[int]]  # List of portal IDs


@dataclass
class SolutionCategory:
    """Solutions: A Freshdesk Category.

    Categories broadly classify your solutions page into several sections.
    """

    id: int  # unique
    name: str
    description: Optional[str]
    visible_in_portals: list[int]  # portal_ids
    created_at: datetime
    updated_at: datetime

    @classmethod
    def from_json(cls, data):
        obj = cls(**data)
        return obj


class SolutionFolderPayload(TypedDict):
    """Used to create or update a solution folder using the API."""

    name: str
    visibility: SolutionFolderVisibility
    description: Optional[str]
    parent_folder_id: Optional[int]
    company_ids: Optional[list[int]]
    contact_segment_ids: Optional[list[int]]
    company_segment_ids: Optional[list[int]]


@dataclass
class SolutionFolder:
    """Solutions: A Freshdesk Folder.

    Related Solutions Articles are organized into Folders.
    """

    id: int  # unique
    category_id: int
    name: str
    description: str
    visibility: SolutionFolderVisibility
    created_at: datetime
    updated_at: datetime
    articles_count: int

    company_ids: list[int] = field(default_factory=list)
    contact_segment_ids: list[int] = field(default_factory=list)
    company_segment_ids: list[int] = field(default_factory=list)

    @classmethod
    def from_json(cls, data):
        obj = cls(**data)
        return obj


class SolutionArticlePayload(TypedDict):
    """Used to create or update a solution article using the API.

    Doesn't _require_ anything when updating an article.
    """

    title: str
    description: str
    status: SolutionArticleStatus
    tags: Optional[list[str]]
    seo_data: Optional[dict]  # meta_title, meta_description, meta_keywords


@dataclass
class SolutionArticle:
    """Solutions: A Freshdesk Article.

    WARNING: There is a large difference between create/update and search
    results.
    """

    id: int  # unique
    type: int
    status: SolutionArticleStatus

    category_id: int
    folder_id: int
    agent_id: int  # Author/Creator

    title: str
    description: str = field(repr=False)
    description_text: str = field(repr=False)

    created_at: datetime

    collaboration: dict = field(default_factory=dict)

    hits: Optional[int] = None
    thumbs_up: Optional[int] = None
    thumbs_down: Optional[int] = None
    suggested: Optional[int] = None
    feedback_count: Optional[int] = None

    # Meta Data for SEO: meta_title, meta_description, meta_keywords
    seo_data: Optional[dict] = None  # noqa: 501
    attachments: Optional[list[Attachment]] = None
    cloud_files: Optional[list] = None

    tags: Optional[list[str]] = None
    folder_visibility: Optional[SolutionFolderVisibility] = None
    path: Optional[str] = None  # Path of the URL (e.g. 123456-article-title)
    modified_at: Optional[datetime] = None  # '2023-03-13T16:09:22Z'
    modified_by: Optional[int] = None  # UserID
    updated_at: Optional[datetime] = None  # '2023-03-13T16:09:22Z'
    language_id: Optional[int] = None  # Language ID
    language: Optional[str] = None  # e.g. 'en'
    category_name: Optional[str] = None
    folder_name: Optional[str] = None

    # Custom
    _json: Optional[dict] = field(default=None, repr=False)

    @classmethod
    def from_json(cls, data):
        copied = deepcopy(data)

        # Parse Datetime strings
        if copied.get("created_at"):
            copied["created_at"] = datetime.fromisoformat(
                copied["created_at"][:-1]
            ).astimezone(timezone.utc)
        if copied.get("updated_at"):
            copied["updated_at"] = datetime.fromisoformat(
                copied["updated_at"][:-1]
            ).astimezone(timezone.utc)
        if copied.get("modified_at"):
            copied["modified_at"] = datetime.fromisoformat(
                copied["modified_at"][:-1]
            ).astimezone(timezone.utc)

        # Convert SolutionArticleStatus
        copied["status"] = SolutionArticleStatus(int(copied["status"]))

        obj = cls(**copied)
        obj._json = data
        return obj


@dataclass
class Contact:
    """A Freshdesk Contact."""

    active: bool
    email: str
    job_title: Optional[str]
    language: str  # Literal['en', ...]
    last_login_at: datetime
    mobile: Optional[str]  # Phone Number
    name: str
    phone: Optional[str]  # Phone Number
    time_zone: str  # Literal['Central Time (US & Canada)', ...]
    created_at: datetime
    updated_at: datetime

    avatar: Optional[dict] = None

    _json: Optional[dict] = field(default=None, repr=False)

    @classmethod
    def from_json(cls, data):
        obj = cls(**data)
        obj._json = data
        return obj


@dataclass
class Agent:
    """A Freshdesk Agent."""

    available: bool
    available_since: datetime
    id: int  # unique
    occasional: bool  # True: occasional, False: full-time
    signature: str = field(repr=False)  # HTML
    ticket_scope: int  # Permissions: 1 - Global, 2 - Group, 3 - Restricted
    # support_agent -> Support Agent
    # field_agent -> Field Agent
    # collaborator -> Collaborator
    type: str
    created_at: datetime
    updated_at: datetime
    last_active_at: datetime
    contact: Contact

    freshcaller_agent: bool

    # Skill IDs
    skill_ids: list[int] = field(default_factory=list)
    # Group IDs
    group_ids: list[int] = field(default_factory=list)
    # Role IDs
    role_ids: list[int] = field(default_factory=list)

    _json: Optional[dict] = field(default=None, repr=False)

    # Any extra arguments get put in here.
    extras: dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Agent:
        copied = deepcopy(data)

        # Parse Data Structures
        copied["contact"] = Contact.from_json(copied["contact"])

        # Remove any keys that are not in the dataclass, putting them into extras
        extras: dict[str, Any] = {}
        for key in list(copied.keys()):
            if key not in cls.__dataclass_fields__:
                extras[key] = copied.pop(key)

        obj = cls(**copied, extras=extras)
        obj._json = data

        return obj

    def __hash__(self):
        return hash(self.id)


@dataclass
class Event:
    """

    Parameters
    ----------
    field_name : str
        Name of the field.
    from_ : str
        Value of the field before the event.
    to : str
        Value of the field after the event.
    """

    field_name: str
    from_: str
    to: str

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Event:
        data = data.copy()
        # We cannot use keywords as attribute names.
        data["from_"] = data.pop("from")
        return cls(**data)

    def to_json(self) -> dict[str, Any]:
        data = self.__dict__
        data["from"] = data.pop("from_")
        return data


class PerformerType(IntEnum):
    """
    enum of PerformerType
    1- Agent
    2- Requester
    3- Agent or Requester
    4- System
    """

    AGENT = 1
    REQUESTER = 2
    AGENT_OR_REQUESTER = 3
    SYSTEM = 4


@dataclass
class Performer:
    """

    Parameters
    ----------
    type : PerformerType
        The type of performer
    members : Optional[list[str]], optional
        Only applicable if type is PerformerType.Agent
        IDs of the agents
    """

    type: PerformerType
    members: Optional[list[str]] = None

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Performer:
        obj = cls(**data)
        return obj

    def to_json(self) -> dict[str, Any]:
        data = self.__dict__
        return data


class ConditionMatchType(Enum):
    ALL = "all"
    ANY = "any"

    def __str__(self) -> str:
        return self.value

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        return super().__eq__(other)


class PropertyResourceType(Enum):
    """Resoure type for a Property for a Condition.

    contacts, tickets, companies, custom_object
    """

    CONTACTS = "contacts"
    TICKETS = "tickets"
    COMPANIES = "companies"
    CUSTOM_OBJECT = "custom_object"

    def __str__(self) -> str:
        return self.value

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        return super().__eq__(other)


class Operator(Enum):
    """An AND/OR operator for a Property.

    True - AND
    False - OR
    """

    AND = True
    OR = False

    def __bool__(self) -> bool:
        return self.value

    def __eq__(self, other):
        if isinstance(other, bool):
            return self.value == other
        return super().__eq__(other)


@dataclass
class Property:
    """Propery of a Condition.

    resource_type : str
        Type of the resource.
    field_name : str
        Name of the field.
    operator : bool
        AND/OR operator to combine multiple conditions in a rule.
    value : str
        Value sest on the selected field.
    object_reference : str
        Ticket's look up field value.
    """

    resource_type: PropertyResourceType
    field_name: str
    operator: Operator
    value: str
    object_reference: str

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Property:
        obj = cls(**data)
        return obj


@dataclass
class Condition:
    name: str
    match_type: ConditionMatchType
    properties: list[Property]


class AutomationRuleActionResourceType(Enum):

    SAME_TICKET = "Same_ticket"
    PARENT_TICKET = "parent_ticket"
    TRACKER_TICKET = "tracker_ticket"
    CUSTOM_OBJECT = "custom_object"

    def __str__(self) -> str:
        return self.value

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        return super().__eq__(other)


class PushToType:
    """Channel which the message should be sent with."""

    SLACK = "Slack"
    OFFICE = "Office365"

    def __str__(self) -> str:
        return self.value  # type: ignore


@dataclass
class AutomationRuleAction:
    field_name: str
    value: str
    email_to: int
    email_body: str
    api_key: str
    auth_header: str
    custom_headers: str
    request_type: str
    url: str
    note_body: str
    notify_agents: list[int]  # Agent IDs
    fwd_to: str
    fwd_cc: str
    fwd_bcc: str
    fwd_note_body: str
    push_to: PushToType
    slack_text: str
    office365_text: str
    resource_type: AutomationRuleActionResourceType
    object_reference: str


class AutomationRuleType(IntEnum):
    TICKET_CREATION = 1
    TIME_TRIGGERS = 3
    TICKET_UPDATES = 4


@dataclass
class AutomationRule:
    """A Freshdesk Automation Rule.

    Parameters
    ----------
    outdated : bool
        Whether the rule is outdated or not. The automation service gets
        changed over time and may make older ruled outdated. These get updated
        after editing and saving.
    last_updated_by : int
        ID of the agent who last updated the rule.
    id : int
        ID of the rule.
    summary : dict
        A summary of the rule. Keys include 'conditions' and 'actions'.
    created_at : datetime
    updated_at : datetime
    affected_tickets_count : int
        Number of tickets affected by the rule, in the given time frame (7
        days).
    meta : dict
        Found when using `view` endpoint.
        total_active_count: int
        total_count: int
    """

    id: int
    position: int
    name: str
    summary: dict[str, Any] = field(repr=False)
    active: bool
    outdated: bool
    last_updated_by: int
    created_at: datetime
    updated_at: datetime
    affected_tickets_count: int
    conditions: list[Condition] = field(repr=False)
    actions: list[AutomationRuleAction] = field(repr=False)

    operator: str = ""
    performer: list[Performer] = field(default_factory=list)
    events: list[str] = field(default_factory=list, repr=False)
    # Some interfaces don't include this
    automation_type_id: Optional[AutomationRuleType] = None
    meta: dict[str, Any] = field(default_factory=dict, repr=False)

    # Extra Fields
    extras: dict[str, Any] = field(default_factory=dict, repr=False)

    _json: Optional[dict] = field(default=None, repr=False)

    @property
    def type(self) -> Optional[AutomationRuleType]:
        """Get the automation type."""
        return self.automation_type_id

    @type.setter
    def type(self, value: AutomationRuleType) -> None:
        """Set the automation type."""
        self.automation_type_id = value

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> AutomationRule:
        copied = deepcopy(data)

        # Parse datetimes
        copied["created_at"] = datetime.fromisoformat(
            copied["created_at"][:-1]
        ).astimezone(timezone.utc)
        copied["updated_at"] = datetime.fromisoformat(
            copied["updated_at"][:-1]
        ).astimezone(timezone.utc)

        # Handle Extras
        extras = {}
        for key in list(copied.keys()):
            if key not in cls.__dataclass_fields__:
                extras[key] = copied.pop(key)

        obj = cls(
            **copied,
            extras=extras,
            _json=data,
        )

        return obj

    def to_json(self) -> dict[str, Any]:
        data = self.__dict__
        # Serialize datetimes
        data["created_at"] = data["created_at"].isoformat()[:-6] + "Z"
        data["updated_at"] = data["updated_at"].isoformat()[:-6] + "Z"
        # Serialize enums
        data["automation_type_id"] = int(data["automation_type_id"])
        # Extras
        for key, value in data.pop("extras").items():
            data[key] = value

        return data


@dataclass
class ScenarioAutomationAction:
    """An action to be performed by an automation rule."""

    name: str
    value: Optional[str] = None


@dataclass
class ScenarioAutomation:
    id: int
    name: str
    description: str
    actions: list[ScenarioAutomationAction] = field(repr=False)
    private: bool

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> ScenarioAutomation:
        # Copy Data to avoid mutating the original
        data = deepcopy(data)

        # Parse Actions
        actions = []
        for action in data["actions"]:
            actions.append(ScenarioAutomationAction(**action))

        obj = cls(**data)
        return obj



"""Limit Information"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Final

from freshdesk.models import Agent
from freshdesk.models import Ticket


ENCODING: Final[str] = "utf-8"


@dataclass
class LimitInfo:
    timestamp: datetime

    calls_per_minute: int
    calls_remaining: int
    calls_consumed: int
    retry_time: int  # seconds



import logging.config

from freshdesk.api import AgentAPI
from freshdesk.api import AutomationAPI
from freshdesk.api import CannedResponseAPI
from freshdesk.api import CannedResponseExtendedAPI
from freshdesk.api import CannedResponseFolderAPI
from freshdesk.api import GroupAPI
from freshdesk.api import ProductAPI
from freshdesk.api import ScenarioAutomationsAPI
from freshdesk.api import SearchResults
from freshdesk.api import SolutionAPI
from freshdesk.api import TicketAPI
from freshdesk.api import TicketFieldAPI
from freshdesk.client import LimitInfo
from freshdesk.client import UnregisteredClient
from freshdesk.constants import MAX_PAGES
from freshdesk.constants import MAX_SEARCH_RESULTS
from freshdesk.constants import RE_FRESHDESK_ARTICLE_URL
from freshdesk.constants import RESULTS_PER_PAGE
from freshdesk.errors import AuthenticationError
from freshdesk.errors import DraftLockedError
from freshdesk.errors import NotFoundError
from freshdesk.logger import get_config
from freshdesk.models import Agent
from freshdesk.models import AutomationRule
from freshdesk.models import AutomationRuleType
from freshdesk.models import CannedResponse
from freshdesk.models import CannedResponseFolder
from freshdesk.models import CannedResponseVisibility
from freshdesk.models import Field
from freshdesk.models import Group
from freshdesk.models import Product
from freshdesk.models import ScenarioAutomation
from freshdesk.models import SolutionArticle
from freshdesk.models import SolutionArticlePayload
from freshdesk.models import SolutionArticleStatus
from freshdesk.models import SolutionCategory
from freshdesk.models import SolutionCategoryPayload
from freshdesk.models import SolutionFolder
from freshdesk.models import SolutionFolderPayload
from freshdesk.models import SolutionFolderVisibility
from freshdesk.models import Ticket
from freshdesk.models import TicketConversation
from freshdesk.models import TicketStatus
from freshdesk.utilities import get_agent_by_name
from freshdesk.utilities import get_date_query
from freshdesk.utilities import get_date_range_query
from freshdesk.utilities import get_remaining_search_results
from freshdesk.utilities import get_ticket_agent


# Logging
logging.config.dictConfig(get_config(__file__))

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


# Register API for Client
Client = UnregisteredClient
Client.register_api(TicketAPI)  # type: ignore
Client.register_api(TicketFieldAPI)  # type: ignore
Client.register_api(ProductAPI)  # type: ignore
Client.register_api(GroupAPI)  # type: ignore
Client.register_api(AgentAPI)  # type: ignore
Client.register_api(CannedResponseAPI)  # type: ignore
Client.register_api(CannedResponseFolderAPI)  # type: ignore
Client.register_api(CannedResponseExtendedAPI)  # type: ignore
Client.register_api(AutomationAPI)  # type: ignore
Client.register_api(SolutionAPI)  # type: ignore
Client.register_api(ScenarioAutomationsAPI)  # type: ignore


__all__ = (
    "Agent",
    "CannedResponse",
    "CannedResponseFolder",
    "CannedResponseVisibility",
    "Client",
    "Field",
    "Group",
    "Product",
    "ScenarioAutomation",
    "Ticket",
    "TicketStatus",
    "TicketConversation",
    "LimitInfo",
    "AutomationRule",
    "AutomationRuleType",
    "RESULTS_PER_PAGE",
    "MAX_PAGES",
    "MAX_SEARCH_RESULTS",
    "RE_FRESHDESK_ARTICLE_URL",
    # Models - Solutions (User Guides)
    "SolutionArticle",
    "SolutionArticlePayload",
    "SolutionArticleStatus",
    "SolutionCategory",
    "SolutionCategoryPayload",
    "SolutionFolder",
    "SolutionFolderPayload",
    "SolutionFolderVisibility",
    # Typing
    "SearchResults",
    # Utilities
    "get_ticket_agent",
    "get_agent_by_name",
    "get_date_query",
    "get_date_range_query",
    "get_remaining_search_results",
    # Exceptions
    "AuthenticationError",
    "DraftLockedError",
    "NotFoundError",
)



from __future__ import annotations

import logging.config
import os
from datetime import datetime
from types import SimpleNamespace
from typing import Final
from typing import Optional

import requests
from requests.models import Response

from freshdesk.base import BaseClient
from freshdesk.enumerators import APIVersion
from freshdesk.enumerators import HTTPRequestMethod
from freshdesk.enumerators import Plan
from freshdesk.enumerators import Resource
from freshdesk.errors import AuthenticationError
from freshdesk.logger import get_config
from freshdesk.limits import LimitInfo


ENCODING: Final[str] = "utf-8"


# Logging
logging.config.dictConfig(get_config(__file__))

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class UnregisteredClient(BaseClient):
    headers = {
        "User-Agent": (
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"
            " AppleWebKit/537.36 (KHTML, like Gecko)"
            " Chrome/122.0.0.0 Safari/537.36"
        )
    }
    api = SimpleNamespace()

    def __init__(
        self,
        domain: str,
        api_key: Optional[str] = None,
        plan: Plan = Plan.BLOSSOM,
        version: APIVersion = APIVersion(2),
    ):
        """Initialize Freshdesk Client.

        Raises
        ------
        AuthenticationError
            If API key is not provided or the environment variable for the API
            key is not set.
        """
        logger.info("Initializing Freshdesk Client.")

        self.domain = domain
        self.plan = plan
        self.version = version

        if api_key is None:
            key = os.environ.get("FRESHDESK_API_KEY")
            if key is None:
                raise AuthenticationError(
                    "API Key is required."
                    " Please set the FRESHDESK_API_KEY environment variable or"
                    " pass the API key as an argument to the `api_key` parameter."
                )
            self.api_key: str = key
        else:
            self.api_key: str = api_key

        self.history: list[Response] = []

        self.limits: list[LimitInfo] = []

    @property
    def hostname(self) -> str:
        return f"https://{self.domain}.freshdesk.com"

    @property
    def base_url(self) -> str:
        return self.hostname + self._api_route + self.version.path

    @property
    def _api_route(self) -> str:
        return "/api"

    def _api_uri(self, resource: Resource) -> str:
        return (
            self.base_url
            + self._api_route
            + self.version.path
            + resource.value
        )

    def _request(self, url: str, method: str = "GET", **kwargs):
        """All HTTP requests are made through this method."""

        if isinstance(method, HTTPRequestMethod):
            method = method.value
        logger.info(f"Making {method} request to {url}.")
        response = requests.request(
            method=method,
            url=url,
            auth=(self.api_key, "dummy"),
            **kwargs,  # type: ignore
        )
        self.history.append(response)
        self._parse_response(response)  # Limits, Pagination, Resource Location

        return response

    @staticmethod
    def _get_limit_info(headers) -> dict[str, int]:
        calls_per_minute = int(headers.get("X-RateLimit-Total", 0))
        calls_remaining = int(headers.get("X-RateLimit-Remaining", 0))
        calls_consumed = int(headers.get("X-RateLimit-Used-CurrentRequest", 0))
        retry_time = int(headers.get("Retry-After", 0))  # seconds

        return dict(
            calls_per_minute=calls_per_minute,
            calls_remaining=calls_remaining,
            calls_consumed=calls_consumed,
            retry_time=retry_time,
        )

    def _parse_response(self, response: requests.Response):
        """Parse response and update self.

        * Limit Information
        * Pagination
        * Resource Location
        """

        headers = response.headers

        # Limit Information
        limit_info: dict[str, int] = self._get_limit_info(headers)
        limits: LimitInfo = LimitInfo(
            timestamp=datetime.now().astimezone(),
            calls_per_minute=limit_info["calls_per_minute"],
            calls_remaining=limit_info["calls_remaining"],
            calls_consumed=limit_info["calls_consumed"],
            retry_time=limit_info["retry_time"],
        )

        # Limit History
        self.limits.append(limits)

        # Pagination, if applicable
        # The 'link' header in the response will hold the next page url if
        # exists. If you have reached the last page of objects, then the link
        # header will not be set.
        pagination_link = headers.get("link", "")
        self.pagination_link = pagination_link  # Hotfix to add this to model

        # Location Header:
        # POST requests will contain the Location Header in the response that
        # points to the URL of the created resource.
        # Response
        # HTTP STATUS: HTTP 201 Created
        # Headers:
        # "Location": https://domain.freshdesk.com/api/v2/tickets/1
        resource_location = headers.get("Location", "")
        self.resource_location = (
            resource_location  # Hotfix to add this to model  # noqa: 501
        )



"""APIs for Freshdesk."""

import logging.config
from typing import Any
from typing import Literal
from typing import Optional
from typing import TypedDict
from typing import Union

from requests import Response
from requests.exceptions import JSONDecodeError

from freshdesk.base import BaseAPI
from freshdesk.base import register_interface
from freshdesk.enumerators import HTTPRequestMethod
from freshdesk.errors import DraftLockedError
from freshdesk.errors import NotFoundError
from freshdesk.logger import get_config
from freshdesk.models import Agent
from freshdesk.models import AutomationRule
from freshdesk.models import AutomationRuleType
from freshdesk.models import CannedResponse
from freshdesk.models import CannedResponseFolder
from freshdesk.models import Field
from freshdesk.models import Group
from freshdesk.models import Product
from freshdesk.models import ScenarioAutomation
from freshdesk.models import SolutionArticle
from freshdesk.models import SolutionArticlePayload
from freshdesk.models import SolutionCategory
from freshdesk.models import SolutionCategoryPayload
from freshdesk.models import SolutionFolder
from freshdesk.models import SolutionFolderPayload
from freshdesk.models import Ticket
from freshdesk.models import TicketConversation
from freshdesk.models import TicketField


Identifier = Union[int, str]


class SearchResults(TypedDict):
    """JSON returned by Ticket Search."""

    total: int
    results: list[Ticket]


# Logging
logging.config.dictConfig(get_config(__file__))

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


ListAllTicketsFilter = Literal[
    "new_and_my_open", "watching", "spam", "deleted"
]  # type alias
ListAllTicketsSort = Literal[
    "created_at", "updated_at", "due_by", "status"
]  # type alias
ListAllTicketsEmbed = Literal[
    "stats", "requester", "description"
]  # type alias


class TicketAPI(BaseAPI):
    name = "tickets"
    path = "/tickets"

    def view(self, id: Identifier):
        """
        # Use 'include' to embed additional details in response
        # Each include will consume an additional credit.

        Embeds
        -----
        * conversations
        * requester
        * company
        * stats - closed_at, resolved_at, first_responded_at times
        """
        url = self.base_url + f"/{id}"
        response = self._request(url=url, method=HTTPRequestMethod.GET)

        try:
            data = response.json()
        except JSONDecodeError:
            raise NotFoundError(f"Ticket #{id} not found.")
        ticket: Ticket = Ticket.from_json(data)

        return ticket

    @register_interface
    def tickets(
        self,
        filter: Optional[ListAllTicketsFilter] = None,
        sort_by: Optional[ListAllTicketsSort] = None,
        ascending: Optional[bool] = None,
        embed: Optional[ListAllTicketsEmbed] = None,
        # Pagination
        page: Optional[int] = None,
        per_page: Optional[int] = None,
    ):
        """List all tickets.

        Parameters
        ----------
        filter : Optional[ListAllTicketsFilter]
            Filter tickets by a predefined filter (per Freshdesk)
        sort_by : Optional[ListAllTicketsSort]
            Sort tickets by a field.
            Default sort order is `created_at` (per Freshdesk)
        ascending : Optional[bool]
            Sort in ascending order.
            Default is descending order (per Freshdesk)
        embed : Optional[ListAllTicketsEmbed]
            Embed additional details in response.
            WARNING: Each include will consume an additional **2** credits.

        Pagination
        ==========
        page : Optional[int]
            The page number to retrieve.
            The default is 1 (per Freshdesk)
        per_page : Optional[int]
            The number of tickets per page.
            The default is 30 (per Freshdesk)
            The max is 100 (per Freshdesk)
        """
        url = self.base_url

        # Options
        if any([filter, sort_by, ascending, embed, page, per_page]):
            url += "?"

        # Options - Base
        if filter:
            url += f"filter={filter}&"
        if sort_by:
            url += f"order_by={sort_by}&"
        if ascending is not None:
            if ascending:
                url += "order_type=asc&"
            else:
                url += "order_type=desc&"
        if embed:
            url += f"include={embed}&"

        # Options - Pagination
        if page:
            url += f"page={page}&"
        if per_page:
            url += f"per_page={per_page}&"

        # Clean Up
        if url.endswith("&"):
            url = url[:-1]

        response = self._request(url, HTTPRequestMethod.GET)

        data = response.json()
        tickets = [Ticket.from_json(ticket_data) for ticket_data in data]

        return tickets

    @register_interface
    def filter_tickets(
        self, query: Optional[Union[str, Field]] = None, page: int = 1
    ) -> SearchResults:
        url = self.client.base_url + "/search" + self.path
        params = {}
        if query is not None:
            # Hotfix for queries already surrounded by quotes
            if str(query)[0] == '"' and str(query)[-1] == '"':
                params = {"query": str(query)}
            else:
                params = {"query": '"' + str(query) + '"'}

        params["page"] = str(page)

        method = HTTPRequestMethod.GET
        response = self._request(url=url, method=method, params=params)

        data: SearchResults = response.json()
        return data

    def list_conversations(self, ticket_id: int) -> list[TicketConversation]:
        url: str = (
            self.client.base_url
            + self.path
            + f"/{ticket_id}"
            + "/conversations"
        )
        response = self._request(method="GET", url=url)
        json_: list[dict[str, Any]] = response.json()
        conversations = [TicketConversation.from_json(c) for c in json_]

        return conversations

    def update_ticket_(self, ticket_id: int, data: dict[str, Any]) -> Ticket:
        """Update a ticket based on ID and data.

        For custom fields, use the following format:

            self.update_ticket(ticket.id, {'custom_fields': {'cf_issue_area': None}})
        """
        url = self.base_url + f"/{ticket_id}"
        method = HTTPRequestMethod.PUT
        headers = {"Content-Type": "application/json"}
        response = self._request(
            url=url, method=method, json=data, headers=headers
        )

        data = response.json()
        data["id"] = ticket_id
        updated_ticket: Ticket = Ticket.from_json(data)

        return updated_ticket

    @register_interface
    def update_ticket(self, ticket_id: int, data: dict[str, Any]) -> Ticket:
        """Update a ticket based on ID and data."""
        return self.update_ticket_(ticket_id, data)

    @register_interface
    def ticket(self, ticket_id: int) -> Ticket:
        logging.info(f"Getting ticket #{ticket_id}.")
        return self.view(ticket_id)


class TicketFieldAPI(BaseAPI):
    name = "ticket_fields"
    path = "/admin/ticket_fields"

    def list_all(self):
        logger.info("Getting all ticket fields.")

        uri = self.base_url
        url = uri
        method = HTTPRequestMethod.GET
        response = self._request(url=url, method=method)
        data = response.json()
        logger.info("Getting all ticket fields.")
        ticket_fields = []
        for field in data:
            logger.info(f"Iterating with {field}.")
            ticket_fields.append(TicketField(**field))

        return ticket_fields

    @register_interface
    def ticket_fields(self) -> list[TicketField]:
        return self.list_all()


class ProductAPI(BaseAPI):
    """API for Products."""

    name = "products"
    path = "/products"

    def list_all(self) -> list[Product]:
        """List all Products."""
        url = self.base_url
        response = self._request(url, HTTPRequestMethod.GET)

        data = response.json()
        products: list[Product] = [
            Product.from_json(product_data) for product_data in data
        ]

        return products


class GroupAPI(BaseAPI):
    """API for Groups."""

    name = "groups"
    path = "/groups"

    def list_all(self) -> list[Group]:
        """List all Groups."""
        url = self.base_url
        response = self._request(url, HTTPRequestMethod.GET)

        data = response.json()
        groups: list[Group] = [
            Group.from_json(group_data) for group_data in data
        ]

        return groups

    @register_interface
    def groups(self) -> list[Group]:
        return self.list_all()


class AgentAPI(BaseAPI):
    """API for Agents."""

    name = "agents"
    path = "/agents"

    def list_all(self) -> list[Agent]:
        """List all Agents."""
        url = self.base_url
        method = HTTPRequestMethod.GET
        response = self._request(method=method, url=url)

        data = response.json()
        groups: list[Agent] = [
            Agent.from_json(agent_data) for agent_data in data
        ]

        return groups

    @register_interface
    def agents(self) -> list[Agent]:
        return self.list_all()


class CannedResponseAPI(BaseAPI):
    """API for Canned Responses."""

    name = "canned_responses"
    path = "/canned_responses"

    def view(self, id: Identifier) -> CannedResponse:
        """View a Canned Response."""
        url = self.base_url + f"/{id}"
        response = self._request(method=HTTPRequestMethod.GET, url=url)

        data = response.json()
        canned_response: CannedResponse = CannedResponse.from_json(data)

        return canned_response

    def update(
        self, id: Identifier, payload: dict[str, Any]
    ) -> CannedResponse:
        """Update a Canned Response."""
        url = self.base_url + f"/{id}"
        response = self._request(
            method=HTTPRequestMethod.PUT, url=url, json=payload
        )

        data = response.json()
        canned_response: CannedResponse = CannedResponse.from_json(data)

        return canned_response

    @register_interface
    def canned_response(self, id: Identifier) -> CannedResponse:
        return self.view(id)

    @register_interface
    def update_canned_response(
        self, id: Identifier, payload: dict[str, Any]
    ) -> CannedResponse:
        return self.update(id, payload)


class CannedResponseFolderAPI(BaseAPI):
    """API for Canned Response Folders."""

    name = "canned_response_folders"
    path = "/canned_response_folders"

    def list_all(self) -> list[CannedResponseFolder]:
        """List all Canned Response Folders."""

        url = self.base_url
        method = HTTPRequestMethod.GET
        response = self._request(method=method, url=url)

        data = response.json()
        canned_response_folders: list[CannedResponseFolder] = [
            CannedResponseFolder.from_json(folder_data) for folder_data in data
        ]

        return canned_response_folders

    def view(self, id: Identifier) -> CannedResponseFolder:
        """View a Canned Response Folder in a simple form."""

        logger.info(f"Getting Canned Response Folder #{id}.")
        url = self.base_url + f"/{id}"
        response = self._request(method=HTTPRequestMethod.GET, url=url)

        data = response.json()
        canned_response_folder: CannedResponseFolder = (
            CannedResponseFolder.from_json(data)
        )

        return canned_response_folder

    def view_canned_responses(self, id: Identifier) -> list[CannedResponse]:
        """View the canned responses in a Canned Response Folder."""

        url = self.base_url + f"/{id}/responses"
        response = self._request(method=HTTPRequestMethod.GET, url=url)

        data = response.json()
        canned_responses: list[CannedResponse] = [
            CannedResponse.from_json(canned_response_data)
            for canned_response_data in data
        ]

        return canned_responses

    @register_interface
    def canned_response_folder(self, id: Identifier) -> CannedResponseFolder:
        return self.view(id)

    @register_interface
    def canned_response_folders(self) -> list[CannedResponseFolder]:
        return self.list_all()

    @register_interface
    def canned_responses(self, folder_id: Identifier) -> list[CannedResponse]:
        return self.view_canned_responses(folder_id)


class CannedResponseExtendedAPI(BaseAPI):
    """Custom extension for the Canned Response and Canned Response Folder APIs.

    Utilizes APIs from CannedResponseAPI and CannedResponseFolderAPI.
    """

    name = "canned_response_extended"

    @register_interface
    def all_canned_responses(self) -> list[CannedResponse]:
        """List all Canned Responses in all Canned Response Folders.

        Uses (1 + n) API calls, where n is the number of Canned Response
        Folders.
        """

        canned_response_folders: list[CannedResponseFolder] = self.client.api.canned_response_folders.canned_response_folders()  # type: ignore
        canned_responses: list[CannedResponse] = []
        folder: CannedResponseFolder
        for folder in canned_response_folders:
            canned_responses += self.client.api.canned_response_folders.canned_responses(folder.id)  # type: ignore

        return canned_responses


class AutomationAPI(BaseAPI):
    """API for Automations."""

    name = "automations"
    path = "/automations"

    def list_all(
        self, type: AutomationRuleType, page: int = 1
    ) -> list[AutomationRule]:
        """List all Automation Rules by type.

        WARNING: There is an undocumented ?page= query parameter that
        paginates the list by 30.
        """

        url = self.base_url + f"/{int(type)}/rules"
        url += f"?page={page}"
        method = HTTPRequestMethod.GET
        response = self._request(method=method, url=url)

        data = response.json()
        rules: list[AutomationRule] = []
        rule_data: dict[str, Any]
        for rule_data in data:
            rule: AutomationRule = AutomationRule.from_json(rule_data)
            # Fix missing type
            rule.type = type
            rules.append(rule)

        return rules

    def view(self, id: Identifier, type: AutomationRuleType) -> AutomationRule:
        """View an Automation Rule."""

        logger.info(f"Getting automation rule for #{id} of type {type}")
        url = self.base_url + f"/{int(type)}/rules/{id}"
        response = self._request(method=HTTPRequestMethod.GET, url=url)

        data = response.json()
        automation_rule: AutomationRule = AutomationRule.from_json(data)
        # Fix missing type
        automation_rule.type = type

        return automation_rule

    def update(
        self, id: Identifier, type: AutomationRuleType, payload: dict[str, Any]
    ) -> AutomationRule:
        """Update an Automation Rule.

        Parameters
        ----------
        id : Identifier
        type : AutomationRuleType
        payload : dict[str, Any]
            These are the changes you'll be making.
            For example, give `{"name": "New Name"}` to change the name.
        """

        logger.info(f"Updating automation rule for #{id} of type {type}")

        url = self.base_url + f"/{int(type)}/rules/{id}"
        method = HTTPRequestMethod.PUT
        response = self._request(method=method, url=url, json=payload)

        if response.status_code != 200:
            raise Exception(
                f"Failed to update automation rule #{id} of type {type}: {response.text}"
            )

        data = response.json()
        automation_rule: AutomationRule = AutomationRule.from_json(data)

        # Fix missing type
        automation_rule.type = type

        return automation_rule

    @register_interface
    def automation(
        self, id: Identifier, type: AutomationRuleType
    ) -> AutomationRule:
        return self.view(id, type)

    @register_interface
    def automations(self) -> list[AutomationRule]:
        """List all automation rules."""
        rules: list[AutomationRule] = []
        for type_ in AutomationRuleType:
            new_rules: list[AutomationRule] = self.list_all(type_)
            rules.extend(new_rules)

            # Paginate
            page: int = 1
            while len(new_rules) == 30:
                page += 1
                new_rules = self.list_all(type_, page=page)
                rules.extend(new_rules)

        return rules

    @register_interface
    def update_automation(
        self, id: Identifier, type: AutomationRuleType, payload: dict[str, Any]
    ) -> AutomationRule:
        return self.update(id, type, payload)


class SolutionAPI(BaseAPI):
    """API for Solutions."""

    name = "solutions"
    path = "/solutions"

    @register_interface
    def create_solution_category(
        self, solution_category_dict: SolutionCategoryPayload
    ) -> SolutionCategory:
        """Create a Solution Category."""

        logger.info(
            "Creating solution category: %s", solution_category_dict["name"]
        )

        url = self.base_url + "/categories"
        method = HTTPRequestMethod.POST
        response = self._request(
            method=method, url=url, json=solution_category_dict
        )

        data: dict = response.json()
        solution_category: SolutionCategory = SolutionCategory.from_json(data)

        return solution_category

    def update_solution_category(
        self, id: Identifier, solution_category_dict: SolutionCategoryPayload
    ) -> SolutionCategory:
        """Update a Solution Category."""
        logger.info(
            "Updating solution category: %s", solution_category_dict["name"]
        )

        url = self.base_url + f"/categories/{id}"
        method = HTTPRequestMethod.PUT
        response = self._request(
            method=method, url=url, json=solution_category_dict
        )

        data: dict = response.json()
        solution_category: SolutionCategory = SolutionCategory.from_json(data)

        return solution_category

    def view_solution_category(self, id: Identifier) -> SolutionCategory:
        """View a Solution Category."""
        logger.info("Viewing solution category: %s", id)

        url = self.base_url + f"/categories/{id}"
        method = HTTPRequestMethod.GET
        response = self._request(method=method, url=url)

        data: dict = response.json()
        solution_category: SolutionCategory = SolutionCategory.from_json(data)

        return solution_category

    def list_all_solution_categories(self) -> list[SolutionCategory]:
        """List all Solution Categories."""

        logger.info("Listing all solution categories")

        url = self.base_url + "/categories"
        method = HTTPRequestMethod.GET
        response = self._request(method=method, url=url)

        data: list[dict] = response.json()
        solution_categories: list[SolutionCategory] = []
        solution_category_data: dict
        for solution_category_data in data:
            solution_category: SolutionCategory = SolutionCategory.from_json(
                solution_category_data
            )
            solution_categories.append(solution_category)

        return solution_categories

    def delete_solution_category(self, id: Identifier) -> None:
        """Delete a Solution Category."""
        logger.info("Deleting solution category: %s", id)

        url = self.base_url + f"/categories/{id}"
        method = HTTPRequestMethod.DELETE
        response = self._request(method=method, url=url)

        if response.status_code != 204:
            raise Exception(
                f"Failed to delete solution category #{id}: {response.text}"
            )

    def create_solution_folder(
        self,
        category_id: Identifier,
        solution_folder_dict: SolutionFolderPayload,
    ) -> SolutionFolder:
        """Create a Solution Folder."""

        logger.info(
            "Creating solution folder: %s", solution_folder_dict["name"]
        )

        url = self.base_url + f"/categories/{category_id}/folders"
        method = HTTPRequestMethod.POST
        response = self._request(
            method=method, url=url, json=solution_folder_dict
        )

        if response.status_code != 201:
            raise Exception(
                f"Failed to create solution folder: {response.text}"
            )

        data: dict = response.json()
        solution_folder: SolutionFolder = SolutionFolder.from_json(data)

        return solution_folder

    def update_solution_folder(
        self, id: Identifier, solution_folder_dict: SolutionFolderPayload
    ) -> SolutionFolder:
        """Update a Solution Folder."""

        logger.info(
            "Updating solution folder: %s", solution_folder_dict["name"]
        )

        url = self.base_url + f"/folders/{id}"
        method = HTTPRequestMethod.PUT
        response = self._request(
            method=method, url=url, json=solution_folder_dict
        )

        data: dict = response.json()
        solution_folder: SolutionFolder = SolutionFolder.from_json(data)

        return solution_folder

    def view_solution_folder(self, id: Identifier) -> SolutionFolder:
        """View a Solution Folder."""
        logger.info("Viewing solution folder: %s", id)

        url = self.base_url + f"/folders/{id}"
        method = HTTPRequestMethod.GET
        response = self._request(method=method, url=url)

        data: dict = response.json()
        solution_folder: SolutionFolder = SolutionFolder.from_json(data)

        return solution_folder

    def list_all_solution_folders(
        self, category_id: Identifier
    ) -> list[SolutionFolder]:
        """List all Solution Folders."""

        logger.info("Listing all solution folders")

        url = self.base_url + f"/categories/{category_id}/folders"
        method = HTTPRequestMethod.GET
        response = self._request(method=method, url=url)

        data: list[dict] = response.json()
        solution_folders: list[SolutionFolder] = []
        solution_folder_data: dict
        for solution_folder_data in data:
            # Fix missing category_id
            solution_folder_data["category_id"] = category_id

            solution_folder: SolutionFolder = SolutionFolder.from_json(
                solution_folder_data
            )
            solution_folders.append(solution_folder)

        return solution_folders

    def list_all_folders_in_folder(
        self, id: Identifier
    ) -> list[SolutionFolder]:
        """List all the Solution Folders found in a Solution Folder."""

        logger.info("Listing all solution folders in folder: %s", id)

        url = self.base_url + f"/folders/{id}/subfolders"
        method = HTTPRequestMethod.GET
        response = self._request(method=method, url=url)

        if response.status_code != 200:
            raise Exception(
                f"Failed to list all solution folders in folder #{id}: {response.text}"
            )

        data: list[dict] = response.json()
        solution_folders: list[SolutionFolder] = []
        solution_folder_data: dict
        for solution_folder_data in data:
            solution_folder: SolutionFolder = SolutionFolder.from_json(
                solution_folder_data
            )
            solution_folders.append(solution_folder)

        return solution_folders

    def delete_solution_folder(self, id: Identifier) -> None:
        """Delete a Solution Folder."""
        logger.info("Deleting solution folder: %s", id)

        url = self.base_url + f"/folders/{id}"
        method = HTTPRequestMethod.DELETE
        response = self._request(method=method, url=url)

        if response.status_code != 204:
            raise Exception(
                f"Failed to delete solution folder #{id}: {response.text}"
            )

    def create_solution_article(
        self,
        folder_id: Identifier,
        solution_article_dict: SolutionArticlePayload,
    ) -> SolutionArticle:
        """Create a Solution Article."""

        logger.info(
            "Creating solution article: %s", solution_article_dict["title"]
        )

        url = self.base_url + f"/folders/{folder_id}/articles"
        method = HTTPRequestMethod.POST
        response = self._request(
            method=method, url=url, json=solution_article_dict
        )

        if response.status_code != 201:
            raise Exception(
                f"Failed to create solution article: {response.text}"
            )

        data: dict = response.json()
        solution_article: SolutionArticle = SolutionArticle.from_json(data)

        return solution_article

    def update_solution_article(
        self, id: Identifier, solution_article_dict: SolutionArticlePayload
    ) -> SolutionArticle:
        """Update a Solution Article."""

        logger.info("Updating solution article: %s", solution_article_dict)

        url = self.base_url + f"/articles/{id}"
        method = HTTPRequestMethod.PUT
        response = self._request(
            method=method, url=url, json=solution_article_dict
        )

        data: dict = response.json()
        # The draft is locked for editing
        if data.get("code") == "draft_locked":
            raise DraftLockedError

        solution_article: SolutionArticle = SolutionArticle.from_json(data)

        return solution_article

    def view_solution_article(self, id: Identifier) -> SolutionArticle:
        """View a Solution Article."""
        logger.info("Viewing solution article: %s", id)

        url = self.base_url + f"/articles/{id}"
        method = HTTPRequestMethod.GET
        response = self._request(method=method, url=url)

        data: dict = response.json()
        solution_article: SolutionArticle = SolutionArticle.from_json(data)

        return solution_article

    def list_all_solution_articles(
        self, folder_id: Identifier
    ) -> list[SolutionArticle]:
        """List all Solution Articles."""

        logger.info("Listing all solution articles")

        url = self.base_url + f"/folders/{folder_id}/articles"
        method = HTTPRequestMethod.GET
        response = self._request(method=method, url=url)

        data: list[dict] = response.json()
        solution_articles: list[SolutionArticle] = []
        solution_article_data: dict
        for solution_article_data in data:
            solution_article: SolutionArticle = SolutionArticle.from_json(
                solution_article_data
            )
            solution_articles.append(solution_article)

        return solution_articles

    def delete_solution_article(self, id: Identifier) -> None:
        """Delete a Solution Article."""
        logger.info("Deleting solution article: %s", id)

        url = self.base_url + f"/articles/{id}"
        method = HTTPRequestMethod.DELETE
        response = self._request(method=method, url=url)

        if response.status_code != 204:
            raise Exception(
                f"Failed to delete solution article #{id}: {response.text}"
            )

    def search_solution_articles(self, keyword: str) -> list[SolutionArticle]:
        """Search for Solution Articles."""

        logger.info("Searching for solution articles: %s", keyword)

        url = self.client.base_url + f"/search/solutions?term={keyword}"
        method = HTTPRequestMethod.GET
        response = self._request(method=method, url=url)

        if response.status_code != 200:
            raise Exception(
                f"Failed to search solution articles: {response.text}"
            )

        data: list[dict] = response.json()
        solution_articles: list[SolutionArticle] = []
        solution_article_data: dict
        for solution_article_data in data:
            solution_article: SolutionArticle = SolutionArticle.from_json(
                solution_article_data
            )
            solution_articles.append(solution_article)

        return solution_articles

    @register_interface
    def create_category(
        self, solution_category_dict: SolutionCategoryPayload
    ) -> SolutionCategory:
        """Create a Solution Category."""
        return self.create_solution_category(solution_category_dict)

    @register_interface
    def update_category(
        self, id: Identifier, solution_category_dict: SolutionCategoryPayload
    ) -> SolutionCategory:
        """Update a Solution Category."""
        return self.update_solution_category(id, solution_category_dict)

    @register_interface
    def category(self, id: Identifier) -> SolutionCategory:
        """Get a Solution Category."""
        return self.view_solution_category(id)

    @register_interface
    def categories(self) -> list[SolutionCategory]:
        """List all Solution Categories."""
        return self.list_all_solution_categories()

    @register_interface
    def delete_category(self, id: Identifier) -> None:
        """Delete a Solution Category."""
        return self.delete_solution_category(id)

    @register_interface
    def create_folder(
        self,
        category_id: Identifier,
        solution_folder_dict: SolutionFolderPayload,
    ) -> SolutionFolder:
        """Create a Solution Folder"""
        return self.create_solution_folder(category_id, solution_folder_dict)

    @register_interface
    def update_folder(
        self, id: Identifier, solution_folder_dict: SolutionFolderPayload
    ) -> SolutionFolder:
        """Update a Solution Folder"""
        return self.update_solution_folder(id, solution_folder_dict)

    @register_interface
    def folder(self, id: Identifier) -> SolutionFolder:
        """Get a Solution Folder."""
        return self.view_solution_folder(id)

    @register_interface
    def folders(self, category_id: Identifier) -> list[SolutionFolder]:
        """List all Solution Folders for a Solution Category."""
        return self.list_all_solution_folders(category_id)

    @register_interface
    def all_folders(self) -> list[SolutionFolder]:
        """List all Solution Folders.

        Uses (1 + category_count) API requests.
        """
        categories: list[SolutionCategory] = self.list_all_solution_categories()  # type: ignore
        folders: list[SolutionFolder] = []
        for category in categories:
            category_folders: list[
                SolutionFolder
            ] = self.list_all_solution_folders(
                category.id
            )  # type: ignore
            folders.extend(category_folders)

        return folders

    @register_interface
    def subfolders(self, id: Identifier) -> list[SolutionFolder]:
        """List all Solution Folders found in a Solution Folder."""
        return self.list_all_folders_in_folder(id)

    @register_interface
    def delete_folder(self, id: Identifier) -> None:
        """Delete a Solution Folder."""
        return self.delete_solution_folder(id)

    @register_interface
    def create_article(
        self,
        folder_id: Identifier,
        solution_article_dict: SolutionArticlePayload,
    ) -> SolutionArticle:
        """Create a Solution Article."""
        return self.create_solution_article(folder_id, solution_article_dict)

    @register_interface
    def update_article(
        self, id: Identifier, solution_article_dict: SolutionArticlePayload
    ) -> SolutionArticle:
        """Update a Solution Article."""
        return self.update_solution_article(id, solution_article_dict)

    @register_interface
    def article(self, id: Identifier) -> SolutionArticle:
        """View a Solution Article."""
        return self.view_solution_article(id)

    @register_interface
    def articles(self, folder_id: Identifier) -> list[SolutionArticle]:
        """List all Solution Articles in an Solution Folder."""
        return self.list_all_solution_articles(folder_id)

    @register_interface
    def all_articles(self) -> list[SolutionArticle]:
        """List all Solution Articles.

        Uses (1 + category_count + all_folders_count) API requests.
        """
        # Gather Folders
        folders: list[SolutionFolder] = self.all_folders()

        # Gather Articles
        articles: list[SolutionArticle] = []
        for folder in folders:
            folder_articles: list[SolutionArticle] = self.list_all_solution_articles(folder.id)  # type: ignore
            articles.extend(folder_articles)

        return articles

    @register_interface
    def delete_article(self, id: Identifier) -> None:
        """Delete a Solution Article."""
        return self.delete_solution_article(id)

    @register_interface
    def search_articles(self, keyword: str) -> list[SolutionArticle]:
        """Search for Solution Articles."""
        return self.search_solution_articles(keyword)


class ScenarioAutomationsAPI(BaseAPI):
    """API for Scenario Automations."""

    name = "scenario_automations"

    def list_all(self) -> list[ScenarioAutomation]:
        """List all Scenario Automations."""

        url: str = self.base_url + "scenario_automations.json"
        method: HTTPRequestMethod = HTTPRequestMethod.GET
        response: Response = self._request(method=method, url=url)

        if response.status_code != 200:
            raise Exception(
                f"Failed to list all scenario automations: {response.text}"
            )

        data: dict = response.json()
        scenario_automations: list[ScenarioAutomation] = []
        scenario_automation_data: dict
        for scenario_automation_data in data:
            scenario_automation: ScenarioAutomation = (
                ScenarioAutomation.from_json(scenario_automation_data)
            )
            scenario_automations.append(scenario_automation)

        return scenario_automations

    @register_interface
    def scenarios(self) -> list[ScenarioAutomation]:
        """List all Scenario Automations."""
        return self.list_all()



import logging.config
import math
from datetime import date
from datetime import datetime
from datetime import timezone
from typing import Optional

import dateutil.parser

from freshdesk.api import SearchResults
from freshdesk.client import UnregisteredClient as FreshdeskClient
from freshdesk.constants import MAX_PAGES
from freshdesk.constants import RESULTS_PER_PAGE
from freshdesk.logger import get_config
from freshdesk.models import Agent
from freshdesk.models import Field
from freshdesk.models import Ticket
from freshdesk.typings import TicketJSON


# YYYY-MM-DDTHH:MM:SSZ
TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"  # UTC


# Logging
logging.config.dictConfig(get_config(__file__))

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def get_agent_by_name(agents: list[Agent], name: str) -> Agent:
    """; Get Agent, given name."""
    agent: Agent
    for agent in agents:
        agent_name: str = agent.contact.name
        if name.lower() in agent_name.lower():
            return agent
    raise Exception(
        f"Agent {name} not found: {[a.contact.name for a in agents]}"
    )


def get_date_query(date_: date) -> str:
    """Get date query, for a specific date.

    Includes updated and created dates.
    """

    date_str: str = date_.strftime(Field.DATE_FORMAT)

    query: str = f"""
    (
        created_at:'{date_str}' OR updated_at:'{date_str}'
    )
    """  # noqa: E501
    query = "".join([line.strip() for line in query.splitlines() if line])

    return query


def get_date_range_query(start_date: date, end_date: date) -> str:
    """Given a start and end date, return a query string for the date range."""

    if start_date == end_date:
        return get_date_query(start_date)

    start_date_str: str = start_date.strftime(Field.DATE_FORMAT)
    end_date_str: str = end_date.strftime(Field.DATE_FORMAT)

    query: str = f"""
    (
        (
            created_at:>'{start_date_str}' AND created_at:<'{end_date_str}'
        ) OR (
            updated_at:>'{start_date_str}' AND updated_at:<'{end_date_str}'
        )
    )
    """  # noqa: E501
    query = "".join([line.strip() for line in query.splitlines() if line])

    return query


def get_remaining_search_results(
    client: "FreshdeskClient",
    query: str,
    search_results: SearchResults,
    strict: bool = False,
) -> list["TicketJSON"]:
    logger.info(f"Retrieving remaining search results: {query=}.")

    # params: dict = {'query': query}
    # url: str = client.base_url + '/search/tickets'

    total: int = int(search_results["total"])
    pages: int = math.ceil(total / RESULTS_PER_PAGE)
    results: list["TicketJSON"] = search_results["results"]

    logger.info(f"There are {pages=} of results.")

    # Whether or not to allow an error with more than available pages.
    if strict:
        page_range = range(2, pages + 1)
    elif not strict:
        page_range = range(2, min(pages + 1, MAX_PAGES))

    for page in page_range:
        logger.info(f"Going to page {page}.")
        # params['page'] = page
        # response = client._request(method='GET', url=url, params=params)
        # sr: SearchResults = response.json()
        sr: SearchResults = client.api.tickets.filter_tickets(
            query=query, page=page
        )

        results.extend(sr["results"])

    return results


def get_ticket_agent(ticket: Ticket, agents: list[Agent]) -> Optional[Agent]:
    """Get the agent associated with the ticket, given all avialable agents."""

    agent: Agent
    for agent in agents:
        if ticket.responder_id == agent.id:
            return agent
    else:
        return None


def parse_timestamp(timestamp: str) -> datetime:
    # timestamp.astimezone(timezone.utc).strftime(TIMESTAMP_FORMAT)
    return datetime.strptime(timestamp, TIMESTAMP_FORMAT).replace(
        tzinfo=timezone.utc
    )


def parse_datetime(dt: str) -> datetime:
    parsed = dateutil.parser.parse(dt)
    if parsed.tzinfo is None:
        parsed = parsed.replace(tzinfo=timezone.utc)
    return parsed
